c
      program adi3d_test
c
c ----------------------------------------------------------------------
c This program does time integration for 3D parabolic systems.
c
c    Author: 
c          Min Xie, April, 98. 
c
c    The equations solved have the form: 
c 
c        Ut = D1Uxx + D2Uyy + D3Uzz +F(U)
c
c BC for diffusing variables:
c Neuman BC:
c        Ux(0,y,z,t)     = G1(y,z,t)
c        Ux(sizex,y,z,t) = G2(y,z,t)
c        Uy(x,0,z,t)     = G3(x,z,t)
c        Uy(x,sizey,z,t) = G4(x,z,t)
c        Uz(x,y,0,t)     = G5(x,y,t)
c        Uz(x,y,sizez,t) = G6(x,y,t)
c
c Dirichlet BC:
c        U(0,y,z,t)     = G1(y,z,t)
c        U(sizex,y,z,t) = G2(y,z,t)
c        U(x,0,z,t)     = G3(x,z,t)
c        U(x,sizey,z,t) = G4(x,z,t)
c        U(x,y,0,t)     = G5(x,y,t)
c        U(x,y,sizez,t) = G6(x,y,t)
c
c    where U = (u1,...,unvar), Di = diag(mu1,..munavr) with mul>=0
c    and x, y, z are spatial variables, standing for the first, 
c    second, and third variable, respectively. The order of the spatial 
c    variables can be changed, up to the user.
c
c   The code makes use of the following 3-d adi scheme. 
c
c    
c   (I-(ht/2)A1)U^* = [I+(ht/2)(A1+2A2+2A3)]U^n + htF(U^n)     step 1
c
c   (I-(ht/2)A2)U^** = U^* - (ht/2)A2U^n                       step 2
c 
c   (I-(ht/2)A3)U^n+1 - (ht/2)f(U^n+1) 
c   = U^** - (ht/2)A3U^n - (ht/2)F(U^n)                        step 2
c
c   where Ai = DiD+D-i and D+D-i denotes the numerical second order 
c   approximation of the second derivative with respect to the i
c   spatial variable.
c
c   In step 1 and 2, a direct tridiagonal matrix solver is used. 
c   In step 3, a nonlinear system of equations is to solve. Three 
c   solvers can be used: 1. nksol, 2. block tridiagonal, 3. tridiagonal. 
c   Note that the second and third solvers are Newton method. In general, 
c   the matrix is block tridiagonal but in the case Jacobian matrix of 
c   F(U) is diagonal, a direct tridiagonal solver can be used. 
c
c ----------------------------------------------------------------------
c
      include 'params.inc'
      include 'nksol.inc'

      real*8 of(meqnz)
      real*8 sf(meqnz)
      real*8 su(meqnz)
      real*8 rwork(lrw)
      integer*4 iwork(liw)

      integer iopt
      integer ipflag
      integer iterm
      integer mdif
      integer mf

      real*8 ftol
      real*8 stptol
c
c Local variables
c
c     ADI variables
c
      integer icol
      integer irow
      integer iplane
      integer ind
      integer index1, index2
c
c  Integrator and looping variables
c
      integer i,j,k,l
      integer istep
      integer idr
c      integer is
      integer iu
      integer ius
      integer jj
c
c  Solution variables Z sweep
c
      real*8 p(3*meqnz+2)    ! zeta 
c
c  Newton solver variables
c
      double precision fp(nvar,nvar), u(nvar)
c
c The solution is stored in the array v.  It is declared
c 4 times as large because we actually need v(t^n+1,t**,t*,t^n))
c
      real*8 v(4*neqn)
c
c  Time and error variables
c
      real*8 tfinal
      integer nsteps
      common /limits/ tfinal, nsteps      
c
      real*8 t
      real*8 tscale   ! min^{-1} time scale
c      real*8 tfinal   ! final time
c      real*8 tslice   ! output time slice
c      real*8 reserr   ! residual error for elliptic solve
      real*8 ht       ! length of time step
      real*8 htx,hty,htz    
      real myclock
      real*8 tfin,tinit
c
c   tridiagonal solver 
c 
      real*8 diagax(meqnx),diagbx(meqnx),diagcx(meqnx)    !sub,main,super 
      real*8 diagay(meqny),diagby(meqny),diagcy(meqny)
      real*8 diagaz(meqnz),diagbz(meqnz),diagcz(meqnz)     
      real*8 diaga(meqn),diagb(meqn),diagc(meqn)          ! pivoting
      real*8 sol(meqn),r_rhs(meqn) 
c
c   non-linear tridiagonal solver 
c 
      integer maxit 
      real*8 fm(meqnz)
c
c   non-linear block tridiagonal solver 
c 
      integer ip(nvar,nvar,nz)
      real*8 blockaz(nvar,nvar,nz)
      real*8 blockbz(nvar,nvar,nz)
      real*8 blockcz(nvar,nvar,nz)
      real*8 blocka(nvar,nvar,nz)
      real*8 blockb(nvar,nvar,nz)
      real*8 blockc(nvar,nvar,nz)
 
c  misc
c
c      logical dynflg
c      logical exflg
c      logical ellflg
      logical nointflg
c      logical outflg
      logical restart
c      logical ssflg
c      logical varstp
c
      character*15 filename1
      character*15 filename2
      character*15 filename3
      character*10 solverx,solvery,solverz
c
      common/ttime/t,ht,htx,hty,htz
      common/solver/solverx,solvery,solverz
c
      external functz
      external jac
      external pset
      external psol
      external myclock
c      
c ****** Begin executable ****** 
c
c initialize parameters etc
c
      open ( unit=60,file='nk.dat',status='unknown' ) 

      call setup(iu,ius,idr,nointflg,restart,tscale,
     .           filename1,filename2,filename3)

      write(iu,*) 't,x,y,info,nl_it, sys_err, residual'

      t=0.d0

      ht = ht0              ! start with default time step
      jj = 0                ! loop counter for main loop
c
c nksol parameters initialized here
c
      iopt = 0               ! nksol user options flag
      iwork(5) = 1           ! print statistics from nksolve
      rwork(3)= 1.d0
      ipflag = 0             ! no preconditioning
      mdif = 0               ! jacobian internally generated
      mf = 3                 ! linear and nonlinear soln method

c
c set the error tolerances for the adi portion of nksol calls
c
      ftol = 1.d-8
      stptol = 0.d0

      write(iu,*)' adi/ftol = ',ftol
      write(iu,*)' adi/stptol = ',stptol
c
c   non-linear tridiagonal solver
c
      maxit = 10
c
c Load the scale factors for p and the function fct
c
      do i = 1,meqnz
         su(i) = 1.0d0
         sf(i) = 1.0d0
      end do
c
c load initial condition (starting guess for elliptic problem)
c
      call start (t,v,restart,idr,filename3)
c---------------------------------------------------------------------
c
c Do any time integration?
c
c     if(nointflg) goto 2
c
c  write initial condition to file
c
      call adiout(t,v,tscale)
c
c  **************************************************************
c
c     Begin main loop:
c
c  ***************************************************************
c
c     This is the return point for the time-stepping loop

 1        jj=jj+1

c  ---------------------------------------------------------------
c
c     Note that htx, etc have to be computed locally when a variable 
c     step size is used
c
        htx = ht/hx**2
        hty = ht/hy**2
        htz = ht/hz**2
c
c  ---------- tridiagonal or block trdiagonal set up ---------------
c
c     Note that we do this within the loop since the time step may be variable
c     par(i+3*(j-1)) contains the diffusion coeffs in the i=x,y,z directions
c     for the jth species. Note that  ht is divided by 2; see algorithm
c--------------------------------------------------------------------
c     for X-direction
c
        call setup_trid(diagax, diagbx, diagcx, nx, meqnx, htx, 1)
c
c     for Y-direction
c
        call setup_trid(diagay, diagby, diagcy, ny, meqny, hty, 2)
c
c     for Z-direction if jacobi of f(u) is a diagonal matrix.
c
        if(solverz .eq. 'tridiag') then
           call setup_trid(diagaz, diagbz, diagcz, nz, meqnz, htz, 3)
        elseif(solverz .eq. 'block_trid') then
           call setup_block_trid(blockaz,blockbz,blockcz,
     .                            nvar,nz,htz,3)

        endif
c-----------------------end tridiagonal setup-----------------------
c
c     Row sweep:   X-direction solve
c
c------------------------------------------------------------------
c
      tinit=myclock()
c
c     Set the  flag istep for the  step in the cycle currently being done. 
c     istep determines which of the four time values is to be replaced in pack
c     istep = 1 -- u*; istep = 2 -- u**; istep = 3 -- u(n+1)
c
c
      istep = 1
c
      do iplane= 1,nz     ! -----------------------------------------
c
       do irow = 1,ny
c
         call rhs_row(irow,iplane,v,r_rhs)
c
         call trid(diagax,diagbx,diagcx,sol,r_rhs,meqnx)
c
c     Update the vector v. We put each new row into u* as it is computed, 
c     rather than doing it all at the end. This is OK, since the old u* is 
c     never used in the computation of the new u*; on;y u^n is used. 
c     
         do l = 1, nvar
            do i=1,nx
               ind = i+(l-1)*nx+(irow-1)*meqnx+(iplane-1)*meqnx*ny
     .              + 2*neqn                        !on u*
               v(ind)=sol(i+(l-1)*nx)
            enddo
         enddo
c       
       end do
      end do             ! -----------------------------------------
c
       tfin=myclock()
c
c--------------------------------------------------------------------------
c     Column sweep:   Y-direction solve
c--------------------------------------------------------------------------

      tinit=myclock()
c
c     Set the  flag that indicates which step in the cycle is being done. 
c
      istep = 2
c
      do iplane = 1,nz          ! -----------------------------------------

c       if(mod(iplane,10).eq.1) print*, 'COLUMN SWEEP for Z=',iplane

       do icol = 1,nx
c
         call rhs_col(icol,iplane,v,r_rhs)
c
         call trid(diagay,diagby,diagcy,sol,r_rhs,meqny)
c
c     Update the vector v. We put each new column into u** as it is computed, 
c     rather than doing it all at the end. This is OK, since the old u* is 
c     never used in the computation of the new u*; on;y u^n is used. 
c     
         do l = 1, nvar
            do j=1,ny
               ind = icol+(l-1)*nx+(j-1)*meqnx+(iplane-1)*meqnx*ny
     .              + neqn                        !on u**
               v(ind)=sol(j+(l-1)*ny)
            enddo
         enddo
c       
       end do
      end do             ! -----------------------------------------
c
c--------------------------------------------------------------------
c     Z-direction solve
c
c  In this step, we solve nonlinear equations
c
c     H(U^n+1) = -(U^n+1 - U^**) + (ht/2)(F(U^n+1)-F(U^n)) = 0
c
c  There are three solvers: 1. nksol, 2. block tridiaganol, 3. tridiagonal
c  Block tridiagonal solver and the direct tridiagonal solver are Newton 
c  solver and the direct tridiagonal solver is used in the case Jacobi of 
c  F is diagonal matrix. 
c--------------------------------------------------------------------
c
c      tinit=myclock()
c      s_t=0.d0 
c
c     Set a flag that indicates which step in the cycle is being done. 
c     This is used to determine how to pack the solution. 
c
      istep = 3
c
      do  irow=1,ny          !--------------------------------------

c         if(mod(irow,10).eq.1) print*, 'ZETA SWEEP for irow =',irow,t
 
         do icol = 1,nx
c-------------------------------------------------------------------
c for fixed icol and irow, put u,u*,oldu into p = (p1,p2,p3,icol,irow) 
c where    
c     p1 = u(icol,irow,(z1,...,znvar)1,...,(z1,...,znvar)nz)
c     p2 = u**(icol,irow,(z1,...,znvar)1,...,(z1,...,znvar)nz)
c     p3 = oldu(icol,irow,(z1,...,znvar)1,...,(z1,...,znvar)nz)
c
           do l = 1, nvar
             do k = 1, nz
                index1 = icol+(l-1)*nx+(irow-1)*meqnx+(k-1)*meqnx*ny
                index2 = l+(k-1)*nvar
c
                p(index2) = v(index1)                !u
                p(index2+meqnz) = v(index1+neqn)     !u**
                p(index2+2*meqnz) = v(index1+3*neqn) !oldu
             enddo
           enddo
           p(3*meqnz+1) = icol              !index for column
           p(3*meqnz+2) = irow              !index for row
c
c--------------------------------------------------------------------
c     Z-direction solve by nksol
c--------------------------------------------------------------------
c
            if(solverz .eq. 'nksol') then
c
            call nksol(meqnz,3*meqnz+2,p,of,functz,jac,su,sf,ftol,
     .           stptol,rwork,lrw,iwork,liw,iopt,iterm,pset,
     .           psol,mf,mdif,ipflag)
            if(iterm .ne. 1) then
               write(6,*) 'Convergence failure at icol,irow=', icol,irow
               goto 666
            endif
c
c--------------------------------------------------------------------
c     Z-direction solve by block tridiag
c--------------------------------------------------------------------
c
         elseif(solverz .eq. 'block_trid') then
c
c     Set a flag that indicates which step in the cycle is being done.
c     This is used to determine how to pack the solution.
c
            call btrsol(blockaz,blockbz,blockcz,p,nvar,nz,maxit,
     .           ftol,t,ht,ip,blocka,blockb,blockc,of,u,fp,sol)
            if(maxit .eq. 10) then
               write(6,*) 'Convergence failure at icol,irow=', icol,irow
               goto 666
            endif
c
c--------------------------------------------------------------------
c     Z-direction solve by tridiag
c--------------------------------------------------------------------
c
         elseif(solverz .eq. 'tridiag') then
c
c     Set a flag that indicates which step in the cycle is being done.
c     This is used to determine how to pack the solution.
c
            call ntrsol(diagaz,diagbz,diagcz,p,nvar,nz,maxit,
     .           ftol,t,ht,diaga,diagb,diagc,fm,of,u,fp,sol)

            if(maxit .eq. 10) then
               write(6,*) 'Convergence failure at icol,irow=', icol,irow
               goto 666
            endif
c
c----------------------------------------------------------------
         endif                      !end z-solve in the loop
c------------------------------------------------------------------
c update v. that is, put p1 into u
c------------------------------------------------------------------
         do l = 1, nvar
            do k = 1, nz
               index1 = icol+(l-1)*nx+(irow-1)*meqnx+(k-1)*meqnx*ny
               index2 = l+(k-1)*nvar
               v(index1) = p(index2)                       !u
            enddo
         enddo
c
      end do
      end do           !---------------------------------------------
c
      tfin=myclock()
c
c---------------------------------------------------------------------------
c
c Put u^n+1, which sits in the first block of v, into u^n, which sits in the 
c last block of v. We could in fact do this directly in ppack!
c
      do i = 1,neqn
         v(i+3*neqn)= v(i)
      end do
c
c     ===============================
c
c     End of time step; update time
c
c     ===============================
c
      t=t+ht

c------------------------------------------------------------------
c     Compare the numerical solution with the exact solution.
c     print out the difference at each points and error in max-norm 
c     and L2 norm at each time.
c------------------------------------------------------------------
c
      call comparison(t,v)
      goto 10
c     
c     output is performed every jn * ht seconds or at the last step
c
c      if(t .ge. tstart .and. mod(jj,jn) .eq. 0)then
c
c         call adiout(t-tstart,v,tscale)
c
c      endif
c
 10   continue
      if(t .lt. tfinal) go to 1      

c *************************************************************** 
c      write(iu,550)
c *************************************************************** 

c     call drop(t,v,idr,filename3)
      stop
c

 666  write(iu,*)' Error  -  iterm not equal to 1'
      write(iu,*)' Iterm = ',iterm
      write(iu,*) 'Program complete'
c
      if(iu.ne.6) close(unit=iu)
      close(unit=ius)
c
      end

c********************************************************************
c     Subroutines
c--------------------------------------------------------------------
c
      subroutine setup_trid(a, b, c, ngrid, n, hts, idir)
      
c--------------------------------------------------------------------
c Setup tridiagonal matrix for tridiagonal solve with boundary condition.
c Input: integer ngrid is the grid point, hts = tau/h^2 where tau is the 
c        time step size and h is space step size, and n is the lenght of 
c        tridiagonal vectors. idir: 1-x, 2-y, 3-z
c Output: three n-vectors a, b, c, stand for sub, diag, super diagonal.
c-----------------------------------------------------------------------
      include 'params.inc'

      integer n, ngrid, l, i, idir
      double precision a(n), b(n), c(n), hts
      character*10 bc

      common/bound_cond/bc
c
c Set a, b, c for non-diffusive variables if there are any.
c
      if(nstat .gt. 0) then
         do  l = 1, nstat
            do  i = 1, ngrid
               b(i+(l-1)*ngrid) = 1.d0    ! main
               a(i+(l-1)*ngrid) = 0.d0    ! sub
               c(i+(l-1)*ngrid) = 0.d0    ! super
            enddo
         enddo
      endif
c-------------------------------------------------------------
c  set a, b, c for diffusive variables with boundary condition.
c--------------------------------------------------------------
c
      do  l = nstat+1, nvar
         do i=1,ngrid
            b(i+(l-1)*ngrid) =  1.d0+hts*par(idir+3*(l-1)) ! main
            a(i+(l-1)*ngrid) = -hts*par(idir+3*(l-1))/2.d0 ! sub
            c(i+(l-1)*ngrid) = a(i+(l-1)*ngrid) ! super
         enddo
            
         a(1+(l-1)*ngrid) = 0.d0
         c(ngrid+(l-1)*ngrid) = 0.d0
c
c     Add the Neumann BC
c
         if(bc .eq. 'neumann') then  
c  
            a(ngrid+(l-1)*ngrid)=a(ngrid+(l-1)*ngrid)*2.d0
            c(1+(l-1)*ngrid)=c(1+(l-1)*ngrid)*2.d0
         endif
c
      enddo

      return
      end

c--------------------------------------------------------------------
c
      subroutine setup_block_trid(a, b, c, nv, nb, hts, idir)
      
c--------------------------------------------------------------------
c Setup block tridiagonal matrix (without Jacobian matrix of F(U^n+1)) 
c for tridiagonal solve with boundary condition.
c
c Input: integer nv is the dimension of each block, hts = tau/h^2 where 
c        tau is the time step size and h is space step size, and nb is 
c        the lenght of block tridiagonal vectors. idir: 1-x, 2-y, 3-z
c Output: three nb-vectors of blocks, a, b, c, stand for sub, diag, super 
c         diagonal blocks.
c-----------------------------------------------------------------------
      include 'params.inc'

      integer nv, nb, i, j, k, idir
      double precision a(nv,nv,nb), b(nv,nv,nb), c(nv,nv,nb), hts
      character*10 bc

      common/bound_cond/bc

      do k = 1, nz
c
c for non-diffusing variables
c
         do i = 1, nstat                  
            do j = 1, nv                 !initialize a,b,c 
               a(i,j,k) = 0.d0
               b(i,j,k) = 0.d0
               c(i,j,k) = 0.d0
            enddo

            b(i,i,k) = -1.d0              !diagonal in each main block 
         enddo
c
c for diffusing variables
c
         do i = nstat+1, nv
            do j = 1, nv                 !initialize a,b,c 
               a(i,j,k) = 0.d0
               b(i,j,k) = 0.d0
               c(i,j,k) = 0.d0
            enddo

            b(i,i,k) = -(1.d0+hts*par(idir+3*(i-1)))!diagonal in main blocks
            a(i,i,k) = 0.5d0*hts*par(idir+3*(i-1))  !diagonal in sub blocks
            c(i,i,k) = a(i,i,k)                     !diagonal in super blocks
         enddo
      enddo
c
c set the first block of the sub and last block of super are zero
c 
      do i = 1, nv
         a(i,i,1) = 0.d0
         c(i,i,nb) = 0.d0
      enddo
c
c Add Neumann BC
c
      if(bc .eq. 'neumann') then
         do i = nstat+1, nv
            a(i,i,nb) = 2.d0*a(i,i,nb)
            c(i,i,1) = 2.d0*c(i,i,1)
         enddo
      endif
         
      return
      end

c-------------------------------------------------------------------------
c     Functions for X, g1: x=0, g2: x=1
c--------------------------------------------------------------------------

      double precision function  g1(l, y, z, t)

      integer l
      double precision   y, z, t

      g1 = 0.d0

      return
      end

c--------------------------------------------------------------------------

      double precision function  g2(l, y, z, t)

      integer l
      double precision   y, z, t

      g2 = 0.d0

      return
      end
c--------------------------------------------------------------------------
c     Functions for Y, g3-y=0, g4-y=1
c--------------------------------------------------------------------------

      double precision function  g3(l, x, z, t)

      double precision   x, z, t

      integer l
      g3 = 0.d0

      return
      end

c--------------------------------------------------------------------------

      double precision function  g4(l, x, z, t)

      integer l
      double precision   x, z, t

      g4 = 0.d0

      return
      end
c--------------------------------------------------------------------------
c     Functions for Z, g5-z=0, g6-z=1
c--------------------------------------------------------------------------

      double precision function  g5(l, x, y, t)

      integer l
      double precision   x, y, t

      g5 = 0.d0

      return
      end

c--------------------------------------------------------------------------

      double precision function  g6(l, x, y, t)

      integer l
      double precision   x, y, t

      g6 = 0.d0

      return
      end
c--------------------------------------------------------------------------

