c--------------------------------------------------------------------
      subroutine  btrsol(a,b,c,p,nv,nb,miter,ftol,t,dt,
     .                ip,blocka,blockb,blockc,of,u,fp,sol)
c--------------------------------------------------------------------
c Block tridiagonal solver is a newton solver that solves nonlinear 
c system of equations:
c            F(U) = 0
c whose Jacobian matrix is block tridiagonal.
c
c Input: nv is the dimension of each block 
c        nb is the lenght block tridiagonal vectors.
c        miter is the maximum number of iterates for Newton method
c        ftol is the error tolerances for the Newton solver
c        p = (p1,p2,p3,irow,icl)
c        t, dt are the time, the time step size.
c        a, b, c, stand for sub, diag, super, three nb-vectors of blocks, 
c        diagonal blocks. They are setuped for diffusing part of the 
c        Jacobian matrix.
c        ip,blocka,blockb,blockc,of,u,fp,sol are used in block 
c        tridiagonal solver.
c Output: p in which p1 is the current solutions
c-----------------------------------------------------------------------

      integer i, j, k, l, iter, length
      integer nv, nb, miter
      integer irow,icol,iplane,ip(nv,nv,nb),ier
      double precision a(nv,nv,nb), b(nv,nv,nb), c(nv,nv,nb)
      double precision blocka(nv,nv,nb)
      double precision blockb(nv,nv,nb)
      double precision blockc(nv,nv,nb)
      double precision p(3*nv*nb+2), ftol, t, dt
      double precision of(nv*nb), res, u(nv)
      double precision fp(nv,nv), sol(nv*nb)

      length = nv*nb              !length of p_i
      icol= idint(p(3*length+1))
      irow= idint(p(3*length+2))

      iter = 0                  ! Initialize the counter for the Newton step
c
c Block tridiagonal for diffusion part.
c
      do k = 1, nb
         do i = 1, nv
            do j = 1, nv
               blocka(i,j,k) = a(i,j,k)
               blockb(i,j,k) = b(i,j,k)
               blockc(i,j,k) = c(i,j,k)
            enddo
         enddo
      enddo
c
c-------------------------------------------------------------------
c Begin the newton loop
c------------------------------------------------------------------
c
 103  call functz(length,3*length+2,p,of)
c
c  Compute the residuals. If the reseidual < ftol, return;
c  otherwise, go on the next iterate.
c
      res = 0.d0
      do l = 1, nv
         do k = 1,nb
            if(dabs(of(l+(k-1)*nv)) .gt. res) then
               res=dabs(of(l+(k-1)*nv))
            endif
         enddo
      enddo

      if(res .lt. ftol) go to 102
c
c get U^n+1 of the iter th iterate and build rhs of Newton method
c ie, -H
c
      do k = 1, nb
         do l=1,nv
            u(l) = p(l+(k-1)*nv)
            sol(l+(k-1)*nv) = -of(l+(k-1)*nv)
         enddo
c
c Compute Jacobian matrix of f at U^n+1 of the iter th iterate
c
         iplane = k
         call funcjac(u,irow,icol,iplane,t+dt,fp)
c
c Add Jacobian matrix of f at U^n+1 of the iter th iterate
c
         do i = 1, nv
            do j = 1, nv
               blockb(i,j,k) = b(i,j,k) + 0.5d0*dt*fp(i,j)
            enddo
         enddo
c
      enddo
c
c     decompose LU 
c
      call  decbt (nv, nb, blockb, blockc, blocka, ip, ier)
c
      if(ier .ne. 0) print*, 'IER=',ier
c
c      solve system 
c 
      call dsolbt(nv,nb,blockb,blockc,blocka,sol,ip)
c    
c     Update the p1 of p
c
      do l = 1, nv
         do k = 1,nb
            p(l+(k-1)*nv) = p(l+(k-1)*nv) + sol(l+(k-1)*nv)
         enddo
      enddo
c
c     If the number of the iterates does not reach the max iterate 
c     number calculate the residual. Otherwise, return.
c
      iter = iter +1
      if(iter .lt. miter) go to 103

 102  miter = iter

      return
      end

c-----------------------------------------------------------------------
      subroutine decbt (m, n, a, b, c, ip, ier)
      integer m, n, ip(m,n), ier
      double precision  a(m,m,n), b(m,m,n), c(m,m,n)
c-----------------------------------------------------------------------
c the following line is for optimized compilation on llnl compilers.
clll. optimize
c-----------------------------------------------------------------------
c block-tridiagonal matrix decomposition routine.
c written by a. c. hindmarsh.
c latest revision.. november 10, 1983 (ach)
c reference.. ucid-30150
c             solution of block-tridiagonal systems of linear
c             algebraic equations
c             a.c. hindmarsh
c             february 1977
c    
c   march 97: single ---> double + LU fact
c              veronese@math.utah.edu

c the input matrix contains three blocks of elements in each block-row,
c including blocks in the (1,3) and (n,n-2) block positions.
c decbt uses block gauss elimination and subroutines dgefa and dgesl
c for solution of blocks.  partial pivoting is done within
c block-rows only.
c
c note.. this version uses linpack routines dgefa/dgesl instead of
c of dec/sol for solution of blocks, and it uses the bla routine ddot
c for dot product calculations.
c
c input..
c     m = order of each block.
c     n = number of blocks in each direction of the matrix.
c         n must be 4 or more.  the complete matrix has order m*n.
c     a = m by m by n array containing diagonal blocks.
c         a(i,j,k) contains the (i,j) element of the k-th block.
c     b = m by m by n array containing the super-diagonal blocks
c         (in b(*,*,k) for k = 1,...,n-1) and the block in the (n,n-2)
c         block position (in b(*,*,n)).
c     c = m by m by n array containing the subdiagonal blocks
c         (in c(*,*,k) for k = 2,3,...,n) and the block in the
c         (1,3) block position (in c(*,*,1)).
c    ip = integer array of length m*n for working storage.
c output..
c a,b,c = m by m by n arrays containing the block lu decomposition
c         of the input matrix.
c    ip = m by n array of pivot information.  ip(*,k) contains
c         information for the k-th digonal block.
c   ier = 0  if no trouble occurred, or
c       = -1 if the input value of m or n was illegal, or
c       = k  if a singular matrix was found in the k-th diagonal block.
c use solbt to solve the associated linear system.
c
c external routines required.. dgefa and dgesl (from linpack) and
c ddot (from the blas, or basic linear algebra package).
c-----------------------------------------------------------------------
         integer nm1, nm2, km1, i, j, k
         double precision  dp, ddot
         if (m .lt. 1 .or. n .lt. 4) go to 210
          nm1 = n - 1
         nm2 = n - 2
c process the first block-row. -----------------------------------------
        call dgefa (a, m, m, ip, ier)
        k = 1
         if (ier .ne. 0) go to 200
        do 10 j = 1,m
           call dgesl (a, m, m, ip, b(1,j,1), 0)
           call dgesl (a, m, m, ip, c(1,j,1), 0)
   10     continue
c adjust b(*,*,2). -----------------------------------------------------
        do 40 j = 1,m
           do 30 i = 1,m
            dp = ddot (m, c(i,1,2), m, c(1,j,1), 1)
            b(i,j,2) = b(i,j,2) - dp
  30       continue
  40     continue
c main loop.  process block-rows 2 to n-1. -----------------------------
        do 100 k = 2,nm1
          km1 = k - 1
          do 70 j = 1,m
           do 60 i = 1,m
              dp = ddot (m, c(i,1,k), m, b(1,j,km1), 1)
              a(i,j,k) = a(i,j,k) - dp
  60         continue
  70       continue
           call dgefa (a(1,1,k), m, m, ip(1,k), ier)
           if (ier .ne. 0) go to 200
         do 80 j = 1,m
  80       call dgesl (a(1,1,k), m, m, ip(1,k), b(1,j,k), 0)
  100    continue
c process last block-row and return. -----------------------------------
        do 130 j = 1,m
           do 120 i = 1,m
            dp = ddot (m, b(i,1,n), m, b(1,j,nm2), 1)
             c(i,j,n) = c(i,j,n) - dp
  120      continue
  130    continue
       do 160 j = 1,m
          do 150 i = 1,m
            dp = ddot (m, c(i,1,n), m, b(1,j,nm1), 1)
             a(i,j,n) = a(i,j,n) - dp
  150      continue
  160    continue
        call dgefa (a(1,1,n), m, m, ip(1,n), ier)
        k = n
       if (ier .ne. 0) go to 200
        return
c error returns. -------------------------------------------------------
  200  ier = k
       return
  210  ier = -1
        return
c-----------------------  end of subroutine decbt  ---------------------
      end
      integer function idamax(n,dx,incx)
c
c     finds the index of element having max. absolute value.
c     jack dongarra, linpack, 3/11/78.
c     modified to correct problem with negative increment, 8/21/90.
c
      double precision dx(1),dmax
      integer i,incx,ix,n
c
      idamax = 0
      if( n .lt. 1 ) return
      idamax = 1
      if(n.eq.1)return
      if(incx.eq.1)go to 20
c
c        code for increment not equal to 1
c
      ix = 1
      if(incx.lt.0)ix = (-n+1)*incx + 1
      dmax = dabs(dx(ix))
      ix = ix + incx
      do 10 i = 2,n
         if(dabs(dx(ix)).le.dmax) go to 5
         idamax = i
         dmax = dabs(dx(ix))
    5    ix = ix + incx
   10 continue
      return
c
c        code for increment equal to 1
c
   20 dmax = dabs(dx(1))
      do 30 i = 2,n
         if(dabs(dx(i)).le.dmax) go to 30
         idamax = i
         dmax = dabs(dx(i))
   30 continue
      return
      end
      subroutine daxpy(n,da,dx,incx,dy,incy)
c
c     constant times a vector plus a vector.
c     uses unrolled loops for increments equal to one.
c     jack dongarra, linpack, 3/11/78.
c
      double precision dx(1),dy(1),da
      integer i,incx,incy,ix,iy,m,mp1,n
c
      if(n.le.0)return
      if (da .eq. 0.0d0) return
      if(incx.eq.1.and.incy.eq.1)go to 20
c
c        code for unequal increments or equal increments
c          not equal to 1
c
      ix = 1
      iy = 1
      if(incx.lt.0)ix = (-n+1)*incx + 1
      if(incy.lt.0)iy = (-n+1)*incy + 1
      do 10 i = 1,n
        dy(iy) = dy(iy) + da*dx(ix)
        ix = ix + incx
        iy = iy + incy
   10 continue
      return
c
c        code for both increments equal to 1
c
c
c        clean-up loop
c
   20 m = mod(n,4)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dy(i) = dy(i) + da*dx(i)
   30 continue
      if( n .lt. 4 ) return
   40 mp1 = m + 1
      do 50 i = mp1,n,4
        dy(i) = dy(i) + da*dx(i)
        dy(i + 1) = dy(i + 1) + da*dx(i + 1)
        dy(i + 2) = dy(i + 2) + da*dx(i + 2)
        dy(i + 3) = dy(i + 3) + da*dx(i + 3)
   50 continue
      return
      end
      double precision function ddot(n,dx,incx,dy,incy)
c
c     forms the dot product of two vectors.
c     uses unrolled loops for increments equal to one.
c     jack dongarra, linpack, 3/11/78.
c
      double precision dx(1),dy(1),dtemp
      integer i,incx,incy,ix,iy,m,mp1,n
c
      ddot = 0.0d0
      dtemp = 0.0d0
      if(n.le.0)return
      if(incx.eq.1.and.incy.eq.1)go to 20
c
c        code for unequal increments or equal increments
c          not equal to 1
c
      ix = 1
      iy = 1
      if(incx.lt.0)ix = (-n+1)*incx + 1
      if(incy.lt.0)iy = (-n+1)*incy + 1
      do 10 i = 1,n
        dtemp = dtemp + dx(ix)*dy(iy)
        ix = ix + incx
        iy = iy + incy
   10 continue
      ddot = dtemp
      return
c
c        code for both increments equal to 1
c
c
c        clean-up loop
c
   20 m = mod(n,5)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dtemp = dtemp + dx(i)*dy(i)
   30 continue
      if( n .lt. 5 ) go to 60
   40 mp1 = m + 1
      do 50 i = mp1,n,5
        dtemp = dtemp + dx(i)*dy(i) + dx(i + 1)*dy(i + 1) +
     *   dx(i + 2)*dy(i + 2) + dx(i + 3)*dy(i + 3) + dx(i + 4)*dy(i + 4)
   50 continue
   60 ddot = dtemp
      return
      end
      subroutine dsolbt (m, n, a, b, c, y, ip)
      integer m, n, ip(m,n)
      double precision  a(m,m,n), b(m,m,n), c(m,m,n), y(m,n)
clll. optimize
c-----------------------------------------------------------------------
c solution of block-tridiagonal linear system.
c coefficient matrix must have been previously processed by decbt.
c m, n, a, b, c, and ip  must not have been changed since call to decbt.
c written by a. c. hindmarsh.
c
c modified: Silvia Veronese march 97 - double precision version
c
c input..
c     m = order of each block.
c     n = number of blocks in each direction of matrix.
c a,b,c = m by m by n arrays containing block lu decomposition
c         of coefficient matrix from decbt.
c    ip = m by n integer array of pivot information from decbt.
c     y = array of length m*n containg the right-hand side vector
c         (treated as an m by n array here).
c output..
c     y = solution vector, of length m*n.
c
c external routines required.. dgesl (linpack) and ddot (blas).
c-----------------------------------------------------------------------
c
      integer nm1, nm2, i, k, kb, km1, kp1
      double precision  dp, ddot
      nm1 = n - 1
      nm2 = n - 2
c forward solution sweep. ----------------------------------------------
      call dgesl (a, m, m, ip, y, 0)
      do 30 k = 2,nm1
        km1 = k - 1
        do 20 i = 1,m
          dp = ddot (m, c(i,1,k), m, y(1,km1), 1)
          y(i,k) = y(i,k) - dp
 20       continue
        call dgesl (a(1,1,k), m, m, ip(1,k), y(1,k), 0)
 30     continue
      do 50 i = 1,m
        dp = ddot (m, c(i,1,n), m, y(1,nm1), 1)
     1     + ddot (m, b(i,1,n), m, y(1,nm2), 1)
        y(i,n) = y(i,n) - dp
 50     continue
      call dgesl (a(1,1,n), m, m, ip(1,n), y(1,n), 0)
c backward solution sweep. ---------------------------------------------
      do 80 kb = 1,nm1
        k = n - kb
        kp1 = k + 1
        do 70 i = 1,m
          dp = ddot (m, b(i,1,k), m, y(1,kp1), 1)
          y(i,k) = y(i,k) - dp
 70       continue
 80     continue
      do 100 i = 1,m
        dp = ddot (m, c(i,1,1), m, y(1,3), 1)
        y(i,1) = y(i,1) - dp
 100    continue
      return
c-----------------------  end of subroutine dsolbt  ---------------------
      end
      SUBROUTINE DGESL(A,LDA,N,IPVT,B,JOB)
      INTEGER LDA,N,IPVT(1),JOB
      DOUBLE PRECISION A(LDA,1),B(1)
C
C     DGESL SOLVES THE DOUBLE PRECISION SYSTEM
C     A * X = B  OR  TRANS(A) * X = B
C     USING THE FACTORS COMPUTED BY DGECO OR DGEFA.
C
C     ON ENTRY
C
C        A       DOUBLE PRECISION(LDA, N)
C                THE OUTPUT FROM DGECO OR DGEFA.
C
C        LDA     INTEGER
C                THE LEADING DIMENSION OF THE ARRAY  A .
C
C        N       INTEGER
C                THE ORDER OF THE MATRIX  A .
C
C        IPVT    INTEGER(N)
C                THE PIVOT VECTOR FROM DGECO OR DGEFA.
C
C        B       DOUBLE PRECISION(N)
C                THE RIGHT HAND SIDE VECTOR.
C
C        JOB     INTEGER
C                = 0         TO SOLVE  A*X = B ,
C                = NONZERO   TO SOLVE  TRANS(A)*X = B  WHERE
C                            TRANS(A)  IS THE TRANSPOSE.
C
C     ON RETURN
C
C        B       THE SOLUTION VECTOR  X .
C
C     ERROR CONDITION
C
C        A DIVISION BY ZERO WILL OCCUR IF THE INPUT FACTOR CONTAINS A
C        ZERO ON THE DIAGONAL.  TECHNICALLY THIS INDICATES SINGULARITY
C        BUT IT IS OFTEN CAUSED BY IMPROPER ARGUMENTS OR IMPROPER
C        SETTING OF LDA .  IT WILL NOT OCCUR IF THE SUBROUTINES ARE
C        CALLED CORRECTLY AND IF DGECO HAS SET RCOND .GT. 0.0
C        OR DGEFA HAS SET INFO .EQ. 0 .
C
C     TO COMPUTE  INVERSE(A) * C  WHERE  C  IS A MATRIX
C     WITH  P  COLUMNS
C           CALL DGECO(A,LDA,N,IPVT,RCOND,Z)
C           IF (RCOND IS TOO SMALL) GO TO ...
C           DO 10 J = 1, P
C              CALL DGESL(A,LDA,N,IPVT,C(1,J),0)
C        10 CONTINUE
C
C     LINPACK. THIS VERSION DATED 08/14/78 .
C     CLEVE MOLER, UNIVERSITY OF NEW MEXICO, ARGONNE NATIONAL LAB.
C
C     SUBROUTINES AND FUNCTIONS
C
C     BLAS DAXPY,DDOT
C
C     INTERNAL VARIABLES
C
      DOUBLE PRECISION DDOT,T
      INTEGER K,KB,L,NM1
C
      NM1 = N - 1
      IF (JOB .NE. 0) GO TO 50
C
C        JOB = 0 , SOLVE  A * X = B
C        FIRST SOLVE  L*Y = B
C
         IF (NM1 .LT. 1) GO TO 30
         DO 20 K = 1, NM1
            L = IPVT(K)
            T = B(L)
            IF (L .EQ. K) GO TO 10
               B(L) = B(K)
               B(K) = T
   10       CONTINUE
            CALL DAXPY(N-K,T,A(K+1,K),1,B(K+1),1)
   20    CONTINUE
   30    CONTINUE
C
C        NOW SOLVE  U*X = Y
C
         DO 40 KB = 1, N
            K = N + 1 - KB
            B(K) = B(K)/A(K,K)
            T = -B(K)
            CALL DAXPY(K-1,T,A(1,K),1,B(1),1)
   40    CONTINUE
      GO TO 100
   50 CONTINUE
C
C        JOB = NONZERO, SOLVE  TRANS(A) * X = B
C        FIRST SOLVE  TRANS(U)*Y = B
C
         DO 60 K = 1, N
            T = DDOT(K-1,A(1,K),1,B(1),1)
            B(K) = (B(K) - T)/A(K,K)
   60    CONTINUE
C
C        NOW SOLVE TRANS(L)*X = Y
C
         IF (NM1 .LT. 1) GO TO 90
         DO 80 KB = 1, NM1
            K = N - KB
            B(K) = B(K) + DDOT(N-K,A(K+1,K),1,B(K+1),1)
            L = IPVT(K)
            IF (L .EQ. K) GO TO 70
               T = B(L)
               B(L) = B(K)
               B(K) = T
   70       CONTINUE
   80    CONTINUE
   90    CONTINUE
  100 CONTINUE
      RETURN
      END
      subroutine  dscal(n,da,dx,incx)
c
c     scales a vector by a constant.
c     uses unrolled loops for increment equal to one.
c     jack dongarra, linpack, 3/11/78.
c     modified to correct problem with negative increment, 8/21/90.
c
      double precision da,dx(1)
      integer i,incx,ix,m,mp1,n
c
      if(n.le.0)return
      if(incx.eq.1)go to 20
c
c        code for increment not equal to 1
c
      ix = 1
      if(incx.lt.0)ix = (-n+1)*incx + 1
      do 10 i = 1,n
        dx(ix) = da*dx(ix)
        ix = ix + incx
   10 continue
      return
c
c        code for increment equal to 1
c
c
c        clean-up loop
c
   20 m = mod(n,5)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dx(i) = da*dx(i)
   30 continue
      if( n .lt. 5 ) return
   40 mp1 = m + 1
      do 50 i = mp1,n,5
        dx(i) = da*dx(i)
        dx(i + 1) = da*dx(i + 1)
        dx(i + 2) = da*dx(i + 2)
        dx(i + 3) = da*dx(i + 3)
        dx(i + 4) = da*dx(i + 4)
   50 continue
      return
      end
      subroutine dgefa(a,lda,n,ipvt,info)
      integer lda,n,ipvt(1),info
      double precision a(lda,1)
c
c     dgefa factors a double precision matrix by gaussian elimination.
c
c     dgefa is usually called by dgeco, but it can be called
c     directly with a saving in time if  rcond  is not needed.
c     (time for dgeco) = (1 + 9/n)*(time for dgefa) .
c
c     on entry
c
c        a       double precision(lda, n)
c                the matrix to be factored.
c
c        lda     integer
c                the leading dimension of the array  a .
c
c        n       integer
c                the order of the matrix  a .
c
c     on return
c
c        a       an upper triangular matrix and the multipliers
c                which were used to obtain it.
c                the factorization can be written  a = l*u  where
c                l  is a product of permutation and unit lower
c                triangular matrices and  u  is upper triangular.
c
c        ipvt    integer(n)
c                an integer vector of pivot indices.
c
c        info    integer
c                = 0  normal value.
c                = k  if  u(k,k) .eq. 0.0 .  this is not an error
c                     condition for this subroutine, but it does
c                     indicate that dgesl or dgedi will divide by zero
c                     if called.  use  rcond  in dgeco for a reliable
c                     indication of singularity.
c
c     linpack. this version dated 08/14/78 .
c     cleve moler, university of new mexico, argonne national lab.
c
c     subroutines and functions
c
c     blas daxpy,dscal,idamax
c
c     internal variables
c
      double precision t
      integer idamax,j,k,kp1,l,nm1
c
c
c     gaussian elimination with partial pivoting
c
      info = 0
      nm1 = n - 1
      if (nm1 .lt. 1) go to 70
      do 60 k = 1, nm1
         kp1 = k + 1
c
c        find l = pivot index
c
         l = idamax(n-k+1,a(k,k),1) + k - 1
         ipvt(k) = l
c
c        zero pivot implies this column already triangularized
c
         if (a(l,k) .eq. 0.0d0) go to 40
c
c           interchange if necessary
c
            if (l .eq. k) go to 10
               t = a(l,k)
               a(l,k) = a(k,k)
               a(k,k) = t
   10       continue
c
c           compute multipliers
c
            t = -1.0d0/a(k,k)
            call dscal(n-k,t,a(k+1,k),1)
c
c           row elimination with column indexing
c
            do 30 j = kp1, n
               t = a(l,j)
               if (l .eq. k) go to 20
                  a(l,j) = a(k,j)
                  a(k,j) = t
   20          continue
               call daxpy(n-k,t,a(k+1,k),1,a(k+1,j),1)
   30       continue
         go to 50
   40    continue
            info = k
   50    continue
   60 continue
   70 continue
      ipvt(n) = n
      if (a(n,n) .eq. 0.0d0) info = n
      return
      end
