      subroutine newton(n, u,res_vec,update_jac,ftol,maxit,ntwork)

c  Purpose- routine to compute a root of f(u) using Newton's
c           method.  No global methods are used here.  Stopping
c           criterion is based on residual error (ftol) and 
c           reaching maximum number of iterations (maxit).
c       
c           This subroutine works in conjuction with adi3d_grid.f 

c  Author- Silvia Veronese
c 
c     Modified by H. OThmer Dec,1996 to remove external reference to 
c         num_update_jac and common grid_stp (which did not match anyway)
c
c  Arguments-
c
c    Input:   n - integer.
c                 order of system to solve.
c             u - double precision vector of length n.
c                 initial guess.
c       res_vec - external
c                 one of the 3 functr,c,p which return F(u) i.e. the residual.
c
c    update_jac - external
c                 routine that builds and update the jac matrix 
c
c           lrw - integer.
c                 length of rw.  solver checks to see if lrw is large
c                 enough.  if not, an error message is placed in a
c                 file named errlog, placed in the current directory.
c                 this message indicates the required size and the
c                 size that was passed in lrw.  unit 99 is used.
c            rw - double precision array of length at least n,
c                 workspace.
c          ftol - double precision
c                 error tolerance for the residual (stopping criterion).
c         maxit - integer
c                 maximum number of newton iterations.
c    Output:  u - double precision vector of length n.
c                 solution 
c         ntwork- ntwork(1) = 1 means ftol was met.
c                           = -1 means ftol not met at maxit iterations.
c                 ntwork(2) = no of nonlinear ite.
c                 ntwork(3) = max system error for tridi solver 
c                 ntwork(4) = residual 
c                 other are undefined up to 10 


      include 'params.inc'

c  Dummy arguments-


      integer n,nmax,maxit,info
      integer lrw
      parameter(nmax=1000)


      double precision u(16*n+2)
      double precision rw(nmax)
      double precision ftol
      double precision stol
      double precision hx,hy,hz,x,pi
      double precision ntwork(10)

c  Local variables-

      integer i
      integer is


      double precision fnorm
      double precision ermax,err,error
      double precision inner
      double precision r(nmax)
c
c  Solver 
c        a,b,c are the sub.main,super diagonals of the jacobian matrix
      double precision a(nmax),b(nmax),c(nmax)
      double precision a1(nmax),b1(nmax),c1(nmax)

c Externals

      external res_vec
      external trid
      external update_jac

c Begin executable

      if(n.gt.nmax) stop 'n too big'
      
      ntwork(1)= 1
      is = 0

      do i=1,n
         rw(i) = 0.0
      end do

c Compute residual:  r = f(u) which is also right hand side of system
c                    according to adi3d splitting 

 1    call res_vec(n, 16*n+2 , u, r)
c     do i=1,n
c     print*, 'res',r(i)
c     enddo

c Test if current u is close enough

      fnorm = 0.0
      do i=1,n
         fnorm = dmax1(fnorm,dabs(r(i)))
      end do

      if(fnorm.lt.ftol) then 
c      print*, 'Number of Iterations:',is
       ntwork(2) = is
       ntwork(4) = fnorm
      return
      endif
      if(is.gt.maxit) then
         print*, 'is.gt.maxit'
         ntwork(1) = -1 
         ntwork(2) = is
         return
      endif

c Update Jacobian

c     print*, 'Iter = ', is ,' Building jacobian, Iter ' 
      call update_jac(u,a,b,c,n)
       do i=1,n
c       write(3,*) a(i),b(i),c(i),u(i)
c       write(3,'(4(e15.5,1x))') a(i),b(i),c(i),u(i)
        a1(i)=a(i)
        b1(i)=b(i)
        c1(i)=c(i)

       enddo 

c Call linear solver: Df(u) s = -f(u)


      call trid(a1,b1,c1,rw,r,n)
      call trid_err2(a,b,c,rw,r, error,n)
c     print*, 'SYS ERR ' ,error
      ntwork(3)=max(ntwork(3),abs(error))


c Compute the Newton update

      do i=1,n
         u(i) = u(i) - rw(i)
      end do


      is = is + 1
c  Check ermax

c     ermax=0.d0
c     pi=4*atan(1.d0)
c     do i=1,n
c      x=(i)*hx
c      err= abs(u(i)-sin(x+pi/6))
c      write(7,'(4(e17.5,1x))') x,u(i),sin(x+pi/6),err
c      ermax =max(ermax,err)
c     enddo
c     print*, 'Iter = ', is ,'ermax=',ermax

      goto 1

      return

      end

